from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

from ansible.module_utils._text import to_bytes, to_text
from ansible.plugins.callback import CallbackBase

from flask import Flask, render_template

import difflib
import os
import jinja2
import json
import subprocess
import threading
import time


app = Flask(__name__)


def mkdiff(d):
    from_lines = d.get('before', '').split('\n')
    from_desc = d.get('before_header', '')
    to_lines = d.get('after', '').split('\n')
    to_desc = d.get('after_header', '')
    return jinja2.Markup(
        difflib.HtmlDiff(wrapcolumn=80).make_table(
            from_lines, to_lines, from_desc, to_desc, True))


@app.route('/')
def homepage():
    changes = []
    for t in app.task_data.itervalues():
        host_changes = dict(filter(
            lambda x: x[1].result.get('changed', False),
            t.host_data.iteritems()))
        if host_changes:
            changes.append({
                'uuid': t.uuid,
                'name': t.name,
                'args': t.args,
                'path': t.path,
                'play': t.play,
                'action': t.action,
                'hosts': host_changes,
            })
    return render_template('index.html',
                           changes=changes,
                           running=app.running,
                           mkdiff=mkdiff)


class TaskData:
    """Data about an individual task."""

    def __init__(self, uuid, name, args, path, play, action):
        self.uuid = uuid
        self.name = name
        self.args = args
        self.path = path
        self.play = play
        self.start = None
        self.host_data = {}
        self.start = time.time()
        self.action = action

    def add_host(self, host):
        if host.uuid in self.host_data:
            if host.status == 'included':
                # concatenate task include output from multiple items
                host.result = '%s\n%s' % (self.host_data[host.uuid].result, host.result)
            else:
                raise Exception('%s: %s: %s: duplicate host callback: %s' % (
                    self.path, self.play, self.name, host.name))

        self.host_data[host.uuid] = host


class HostData:
    """Data about an individual host."""

    def __init__(self, uuid, name, status, result):
        self.uuid = uuid
        self.name = name
        self.status = status
        self.result = result
        self.finish = time.time()


class CallbackModule(CallbackBase):

    CALLBACK_VERSION = 2.0
    CALLBACK_TYPE = 'aggregate'
    CALLBACK_NAME = 'webdiff'
    CALLBACK_NEEDS_WHITELIST = True

    def __init__(self):
        super(CallbackModule, self).__init__()

        self._playbook_path = None
        self._playbook_name = None
        self._play_name = None
        self._task_data = {}

        self._start_app()

    def _start_app(self):
        app.config.update({'DEBUG': False})
        app.task_data = self._task_data
        app.running = True
        t = threading.Thread(target=app.run,
                             args=('127.0.0.1', 3434))
        t.setDaemon(True)
        t.start()

        print('Started HTTP server on http://127.0.0.1:3434/')
        if os.getenv('DISPLAY'):
            subprocess.call(['xdg-open', 'http://127.0.0.1:3434/'])

    def _start_task(self, task):
        """Record the start of a task for one or more hosts."""
        uuid = task._uuid
        if uuid in self._task_data:
            return

        play = self._play_name
        name = task.get_name().strip()
        path = task.get_path()
        action = task.action

        args = {}
        if not task.no_log:
            args = task.args

        self._task_data[uuid] = TaskData(uuid, name, args, path, play, action)

    def _finish_task(self, status, result):
        """Record the results of a task for a single host."""

        task_uuid = result._task._uuid

        if hasattr(result, '_host'):
            host_uuid = result._host._uuid
            host_name = result._host.name
        else:
            host_uuid = 'include'
            host_name = 'include'

        if hasattr(result, '_result'):
            result = result._result

        task_data = self._task_data[task_uuid]
        task_data.add_host(HostData(host_uuid, host_name, status, result))

    def v2_playbook_on_start(self, playbook):
        self._playbook_path = playbook._file_name
        self._playbook_name = os.path.splitext(
            os.path.basename(self._playbook_path))[0]

    def v2_playbook_on_play_start(self, play):
        self._play_name = play.get_name()

    def v2_runner_on_no_hosts(self, task):
        self._start_task(task)

    def v2_playbook_on_task_start(self, task, is_conditional):
        self._start_task(task)

    def v2_playbook_on_cleanup_task_start(self, task):
        self._start_task(task)

    def v2_playbook_on_handler_task_start(self, task):
        self._start_task(task)

    def v2_runner_on_failed(self, result, ignore_errors=False):
        self._finish_task('failed', result)

    def v2_runner_on_ok(self, result):
        self._finish_task('ok', result)

    def v2_runner_on_skipped(self, result):
        self._finish_task('skipped', result)

    def v2_playbook_on_include(self, included_file):
        self._finish_task('included', included_file)

    def v2_playbook_on_stats(self, stats):
        app.running = False

        # Dump the task data for debugging.
        # for task_data in self._task_data.values():
        #     print('task %s/%s' % (task_data.name, task_data.action))
        #     for host_data in task_data.host_data.values():
        #         result_str = '%s' % host_data.result
        #         if len(result_str) > 100:
        #             result_str = result_str[:100] + '...'
        #         print('  %s -> %s, %s' % (
        #             host_data.name, host_data.status, result_str))

        raw_input('\nPress Enter to stop the web server and exit')

